var toolset 	= require('toolset');
var prompt 		= require('prompt');
var colors 		= require('colors');
var _ 			= require('underscore');
var targz 		= require('tar.gz');
var lingo 		= require('lingo');
var path		= require('path');
var qs			= require('querystring');
var fs			= require('fs');
var request		= require('request');
var formData 	= require('form-data');
var http 		= require('http');
http.post 		= require('http-post');
//var progressbar	= require('pace');
var progressbar = require('progress');
var os 			= require("os");
var mkdirp 		= require("mkdirp");

String.prototype.replaceAll = function(find, replace) {
	return this.replace(new RegExp(find, 'g'), replace);
}


colors.setTheme({
	silly: 'rainbow',
	input: 'grey',
	verbose: 'cyan',
	prompt: 'grey',
	info: 'green',
	data: 'grey',
	help: 'cyan',
	warn: 'yellow',
	debug: 'blue',
	error: 'red'
});


var pagevamp = function() {
	this.env			= 'local';
	this.file = {	// Credential file set by the loadSettings
		settings:		path.normalize(path.join(__dirname,'..','data','storage'))+path.sep+'settings.dat'
	};
	this.scaffoldDirectory	= path.normalize(path.join(__dirname,'..','data','scaffolds'))+path.sep;
	this.buildDirectory		= os.tmpdir()+path.sep;
	this.apibase			= "http://localhost:5020/api/";
}


pagevamp.prototype.init = function(callback) {
	var scope = this;
	this.asciiart = {
		pagevamp:	[
			" _ __   __ _  __ _  _____   ____ _ _ __ ___  _ __  ",
			"| '_ \\ / _` |/ _` |/ _ \\ \\ / / _` | '_ ` _ \\| '_ \\ ",
			"| |_) | (_| | (_| |  __/\\ V / (_| | | | | | | |_) |",
			"| .__/ \\__,_|\\__, |\\___| \\_/ \\__,_|_| |_| |_| .__/ ",
			"|_|          |___/                          |_|    ",
			"                                     version "+this.version+" "
		],
		files:	[
			"  __ _ _           ",
			" / _(_) | ___  ___ ",
			"| |_| | |/ _ \\/ __|",
			"|  _| | |  __/\\__ \\",
			"|_| |_|_|\\___||___/",
			"                   "
		],
		status:	[
			"     _        _             ",
			" ___| |_ __ _| |_ _   _ ___ ",
			"/ __| __/ _` | __| | | / __|",
			"\\__ \\ || (_| | |_| |_| \\__ \\",
			"|___/\\__\\__,_|\\__|\\__,_|___/",
			"                            "
		]
	};
	// Check the local settings
	this.reloadSettings(function() {
		scope.file.credentials	= path.normalize(path.join(__dirname,'..','data','storage'))+path.sep+'profile-'+scope.env+'.dat';
		if (callback) {
			callback();
		}
	});
};


pagevamp.prototype.reloadSettings = function(callback) {
	var scope = this;
	this.getSettings(function(settings) {
		var i;
		for (i in settings) {
			scope[i] = settings[i];
		}
		if (callback) {
			callback();
		}
	});
};


pagevamp.prototype.displayArt = function(name, color) {
	if (!color) {
		color = 'help';
	}
	_.each(this.asciiart[name], function(line) {
		console.log(line[color]);
	});
}


pagevamp.prototype.getSettings = function(callback) {
	var scope = this;
	toolset.file.exists(this.file.settings, function(exists) {
		if (exists === true) {
			toolset.file.toObject(scope.file.settings, function(obj) {
				if (!obj) {
					callback({});
				} else {
					callback(obj);
				}
			});
		} else {
			callback({});
		}
	});
}
pagevamp.prototype.setSettings = function(data, callback) {
	var scope = this;
	scope.updateEncrypted(scope.file.settings, data, callback);
}

pagevamp.prototype.getLoginData = function(callback) {
	var scope = this;
	toolset.file.exists(this.file.credentials, function(exists) {
		if (exists === true) {
			toolset.file.toObject(scope.file.credentials, function(obj) {
				if (!obj || !obj.apikey || !obj.secret) {
					scope.login(callback);
				} else {
					console.log("\n\nWelcome back, "+obj.firstname+"!\n\n".info);
					callback(obj);
				}
			});
		} else {
			// No credentials saved, we ask for the user to login
			scope.login(callback);
		}
	});
}


pagevamp.prototype.createProject = function(type, name, link) {	// link: theme's ID, if it's a sub-component
	var scope = this;
	
	var shortName = lingo.camelcase(name);
	
	// get the login data
	this.getLoginData(function(loginData) {
		
		// Make sure the scaffold exists
		toolset.file.exists(scope.scaffoldDirectory+type+'.tar.gz', function(exists) {
			if (!exists) {
				console.log("That scaffold doesn't exist. Did you type it right?".error);
				return false;
			}
			
			console.log("Creating project".data, shortName.info);
			var directoryName = scope.directory+'/'+shortName;
			
			toolset.archive.extract(scope.scaffoldDirectory+type+'.tar.gz', directoryName, function() {
				console.log("\n\nYour project ".info+"".error+" has been created in ".info+path.normalize(directoryName));
				console.log("Please wait while we setup the project's data...".data);
				
				// Now we check if there is a valid project in that directory
				scope.getProjectData(function(data) {
					if (!data) {
						console.log("");
						console.log("Warning!".error);
						console.log("The file component.json is either missing or invalid.".error);
						console.log("You won't be able to test and publish your app to Pagevamp.".data);
						return false;
					}
					
					// Update the project file
					data.name 			= shortName;
					data.displayName 	= name;
					data.id 			= 'pv'+toolset.crypto.uuid();
					
					// If it's a subcomponent of a theme, link it here
					if (link) {
						data.link		= link;
					}
					
					var stack = new toolset.stack();
					
					// Save the project data
					stack.add(function(p, cb) {
						scope.updateProjectData(data, cb, path.normalize(directoryName));
					});
					
					// Update the project files
					// Update the directive
					if (data.directive) {
						stack.add(function(p, cb) {
							// Read the directive
							var filename = path.normalize(directoryName+path.sep+data.directive);
							toolset.file.read(filename, function(filedata) {
								if (filedata) {
									filedata = filedata.replaceAll('%id%', 			data.id);
									filedata = filedata.replaceAll('%displayname%', data.displayName);
									filedata = filedata.replaceAll('%name%', 		data.name);
									// Save the data
									toolset.file.write(filename, filedata, cb);
								} else {
									cb();
								}
							});
						});
					}
					if (data.editor) {
						stack.add(function(p, cb) {
							// Read the directive
							var filename = path.normalize(directoryName+path.sep+data.editor);
							toolset.file.read(filename, function(filedata) {
								if (filedata) {
									filedata = filedata.replaceAll('%id%', 			data.id);
									filedata = filedata.replaceAll('%displayname%', data.displayName);
									filedata = filedata.replaceAll('%name%', 		data.name);
									// Save the data
									toolset.file.write(filename, filedata, cb);
								} else {
									cb();
								}
							});
						});
					}
					
					stack.process(function() {
						console.log("Project updated.".data);
					});
					
				}, directoryName);
			});
		});
		
	});
}

pagevamp.prototype.login = function(callback, message) {
	var scope = this;
	
	prompt.start()
	prompt.message = "";
	prompt.delimiter = "";
	
	if (!message) {
		message = [
			"",
			"You are not logged in yet.".info,
			"Please enter your developer key and developer secret.",
			"You will find those keys under your account, on https://apps.pagevamp.com",
			"",
			"We will only ask you for those keys once.",
			""
		];
	}
	
	_.each(message, function(msg) {
		console.log(msg);
	});
	
	var prompts = {
		properties: {
			key: {
				pattern: /^[a-zA-Z0-9]+$/,
				message: 'Invalid key format.',
				description: "API KEY:",
				required: true
			},
			secret: {
				pattern: /^[a-zA-Z0-9]+$/,
				message: 'Invalid key format.',
				description: "SECRET:",
				required: true
			}
		}
	};
	
	prompt.get(prompts, function (err, result) {
		if (result && result.key && result.secret) {
			// Try to login
			scope.verifyCredentials(result, function(response){
				if (!response) {
					// Credentials do not match :(
					/// Try to login again
					scope.login(callback, [
						'',
						'Oops!'.error,
						'The key and/or secret you entered do not match any developer account.',
						''
					]);
				} else {
					// All good!
					// Let's greet the user
					console.log("\nWelcome, "+response.firstname+"!\n".info);
					// Let's save that and keep going
					scope.writeEncrypted(scope.file.credentials, response, callback);
				}
			});
		} else {
			console.log("Canceled".error);
		}
		
	});
}


pagevamp.prototype.writeEncrypted = function(filename, data, callback) {
	toolset.file.write(filename, JSON.stringify(data, null, 4), function() {
		callback(data);
	});
}
pagevamp.prototype.updateEncrypted = function(filename, data, callback) {
	var scope = this;
	toolset.file.exists(filename, function(exists) {
		if (exists === true) {
			toolset.file.toObject(filename, function(obj) {
				if (!obj) {
					// Create
					scope.writeEncrypted(filename, data, callback);
				} else {
					scope.writeEncrypted(filename, _.extend(obj, data), callback);
				}
			});
		} else {
			// Create
			scope.writeEncrypted(filename, data, callback);
		}
	});
	
}


pagevamp.prototype.verifyCredentials = function(credentials, callback) {
	//@todo: When we have an app store
	
	if (credentials.key == 'test' && credentials.secret == 'test') {
		callback({
			apikey:		credentials.key,
			secret:		credentials.secret,
			firstname:	'Test Account',
			lastname:	'Demo',
			gender:		'bot'
		});
		return false;
	}
	
	this.apicall({
		method:		'user/auth',
		params:		{
			apikey:		credentials.key,
			secret:		credentials.secret
		}
	}, function(response) {
		if (!response) {
			if (response.error && response.error.message && response.error.message.error) {
				console.log(response.error.message.error);
			} else {
				console.log("Unable to connect to the server. Please try again later.");
			}
			
		} else {
			callback(response);
		}
	});
	
	
}


pagevamp.prototype.apicall = function(options, callback) {
	options = _.extend({
		method:		'',
		params:		{}
	}, options);
	
	toolset.file.toObject(this.apibase+options.method+'?'+qs.stringify(options.params), function(response) {
		callback(response);
	});
}


pagevamp.prototype.logout = function(callback) {
	toolset.file.removeFile(this.file.credentials, callback);
}


pagevamp.prototype.status = function() {
	
}


pagevamp.prototype.stash = function(name) {
	var zipfile = path.normalize(this.scaffoldDirectory+lingo.camelcase(name)+'.tar.gz');
	toolset.archive.compressDirectory(this.directory, zipfile, function(zipfile) {
		console.log("Your working directory has been saved a a template.".data);
		console.log("To create a new project using that template, type:".data);
		console.log("\tpagevamp create ".info+lingo.camelcase(name).info);
		console.log("zipfile",zipfile);
	});
}


pagevamp.prototype.publish = function(versionType) {
	var scope = this;
	
	this.displayArt('pagevamp');
	this.getLoginData(function(user) {
		// Increment the version
		scope.incrementVersion(versionType, false, function(version) {
			var versionStr 	= [version.major, version.minor, version.build].join('.');
			console.log("Version: ", versionStr.info);
			console.log("packing... Please wait...".data);
			console.log(("(packing in "+scope.buildDirectory+")").data);
			scope.pack(versionStr, function(zipfile) {
			console.log(("Package created as "+zipfile).data);
				// Zip is created.
				// Now we upload it to the store
				console.log('');
				console.log('Uploading the component, please wait...'.data);
				/*
				fs.watchFile(zipfile, function () {
					fs.stat(zipfile, function (err, stats) {
						console.log('state', stats.size);
					});
				});*/
				
				// Get the size of the zipfile
				fs.stat(zipfile, function (err, stats) {
					var zipSize			= stats.size;
					var uploadedSize	= 0; // Incremented by on('data') to keep track of the amount of data we've uploaded
					
					// Create a new read stream so we can plug events on it, and get the upload progress
					var zipReadStream 	= fs.createReadStream(zipfile);
					
					//var progress		= progressbar(zipSize);
					var bar = new progressbar('  Uploading [:bar] :percent :etas', {
						complete: 	'=',
						incomplete:	' ',
						width: 		20,
						total: 		zipSize
					});
					
					zipReadStream.on('data', function(buffer) {
						var segmentLength 	= buffer.length;
						uploadedSize 		+= segmentLength;
						bar.tick(segmentLength);
						//console.log("Progress:\t",((uploadedSize/zipSize*100).toFixed(2)+"%").data);
						//progress.op(segmentLength);
					});
					zipReadStream.on('end', function() {
						//console.log("Event: end");
					});
					zipReadStream.on('close', function() {
						//console.log("Event: close");
						console.log('');
						console.log('Upload completed.'.data);
						console.log('The app server is now verifying the data... Please wait...'.data);
					});
					//@TODO: bug when first login. user is null.
					var form = new formData();
					form.append('apikey', user.apikey);
					form.append('secret', user.secret);
					form.append('component', zipReadStream);
					form.submit(scope.apibase+'cli/upload', function(err, res) {
						if (err) {
							console.log("Oups! We encountered an error :(\n\n".error, err);
							return false;
						}
						console.log('');
						console.log("Your component has been published.".info);
						
						var body = '';
						
						res.on('data', function (chunk) {
							body += chunk;
						});
						res.on('end', function () {
							
							// Convert to response to JSON
							try {
								body = JSON.parse(body);
							} catch (e) {
								console.log('An unknown error occured:'.red);
								console.log(body);
								res.resume()
								return false;
							}
							
							// Check for errors
							if (body.error) {
								console.log('An error has occured:'.red);
								console.log(body.error.message.red);
								res.resume()
								return false;
							}
							
							// We're all good!
							res.resume()
						});
						
						
						;
					});
				});
				
				return false;
				
				// DFW
			});
		});
	});
	
	
}


// pack the current project
pagevamp.prototype.pack = function(version, callback) {
	
	var scope = this;
	
	this.getProjectData(function(data) {
		
		var buildDir = path.normalize(scope.buildDirectory+data.id+path.sep);
		
		// Create the tmp build dir path
		mkdirp(buildDir, function(err) {
			if (err) console.error(err);
			// Zip the directory
			
			var filename = buildDir + version + '-' + data.id + '.tar.gz';
			
			toolset.archive.compressDirectory(scope.directory, filename, function(zipfile) {
				callback(filename);
			});
		});
	});
}


// Increment the version of a project
pagevamp.prototype.incrementVersion = function(type, directory, cb) {
	var scope = this;
	this.getProjectData(function(data) {
		data.version = _.extend({
			major:	0,
			minor:	0,
			build:	0
		}, data.version);
		
		if (!data.version[type] && data.version[type] !== 0) {
			type = 'build';
		}
		
		// Reset lower version types
		if (type == 'major') {
			data.version.minor = 0;
			data.version.build = 0;
		}
		if (type == 'minor') {
			data.version.build = 0;
		}
		
		// Increment the version
		data.version[type]++;
		
		// Save the version
		scope.updateProjectData(data, function() {
			cb(data.version);
		}, directory);
	});
}


// Get the current project's data
pagevamp.prototype.getProjectData = function(callback, directory, silent) {
	var scope = this;
	
	if (!directory) {
		directory = this.directory;
	}
	var filename = path.normalize(directory+path.sep+"component.json");
	
	toolset.file.exists(filename, function(exists) {
		if (!exists && !silent) {
			console.log("");
			console.log("This doesn't seem to be a valid Pagevamp project.".error);
			console.log("the file ".error+"component.json".warn+" is not present in your current directory.".error);
			console.log("Use"+" pagevamp create [type]".data+" to create a new project.");
			return false;
		}
		
		toolset.file.toObject(filename, function(data) {
			if (!data && !silent) {
				console.log("");
				console.log("Your component.json file is not valid JSON. Please fix that file and retry.".error);
				console.log("Filename: ",filename);
				console.log("data: ",data);
				callback(false);
				return false;
			}
			callback(data);
		});
	});
}


// Get the current project's data
pagevamp.prototype.updateProjectData = function(data, callback, directory) {
	var scope = this;
	
	if (!directory) {
		directory = this.directory;
	}
	var filename = directory+path.sep+"component.json";
	
	toolset.file.write(filename, JSON.stringify(data, null, 4), function() {
		callback();
	});
}


module.exports = new pagevamp();